/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef HTN
#define HTN

#include <map>
#include <list>

/* HTN weighted hyperarc data type definition */
struct t_hyperarc{
	unsigned long idx;
	unsigned long tail;
	std::map<unsigned long, long> weighted_heads;	
};

/* HTN feasible scheduling data type definition */
struct t_feasible_sched{
	unsigned long size;
	long* p;
};

/* HTN negative cycle data type definition */
struct t_neg_cycle{
	std::map<unsigned long, t_hyperarc*> SC;
};

/* Hyper Temporal Network definition */
class HyperTemporalNetwork{
	private:
	unsigned long n; // number of nodes
	unsigned long e; // number of hyper-arcs
	unsigned long m; // number of internal arcs
	std::list<t_hyperarc*>* hyperarcs; // array of hyperarcs lists  
	
	void load(const char* f);
	void init_hyperarcs();
	void delete_hyperarcs();
	
	public:
	HyperTemporalNetwork(const char* f);
	HyperTemporalNetwork(unsigned long n); 
	~HyperTemporalNetwork();
	
	unsigned long get_n();
	unsigned long get_e();
	unsigned long get_m();
	std::list<t_hyperarc*>* get_hyperarcs(unsigned long h);
	void push_hyperarc(unsigned long h, t_hyperarc* e);
	bool is_well_defined();
	void assert_feasible_sched(t_feasible_sched* t);
	void assert_neg_cycle(t_neg_cycle* c);
	void print();
};

#endif
