/* 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* 
    This code is an implementation of the Value Iteration Algorithm for Mean Payoff Energy Games 
     see [Brim2011]: 
       L. Brim et. al,  Journal Formal Methods in System Design, Volume 38 Issue 2, April 2011, Pages 97-118 
                        Kluwer Academic Publishers Hingham, MA, USA 
*/

#ifndef BRIM
#define BRIM

#include <list>
#include <utility>
#include "../../h/mpg.h"
#include "../../../h/conf.h"

class Brim2011Algorithm{
	private:
	MeanPayoffGame* mpg;
	unsigned long M_G;

	unsigned long get_M_G();
	unsigned long circle_op(long a, long b);
	void lift_op(unsigned long* e, unsigned long v);
	std::list<std::pair<unsigned long, long> >* compute_pre_arcs();
	long get_count(unsigned long* e, unsigned long v);
	t_w_arc* compute_max_winning_strategy(MeanPayoffGame* mpg);	
	std::list<t_w_arc>* invert_arc_list(std::list<t_w_arc>* arc_list);
	t_w_arc* invert_strategy(t_w_arc* pi);
	MeanPayoffGame* invert_mpg();

	public:
	Brim2011Algorithm(MeanPayoffGame *mpg);
	~Brim2011Algorithm();
	
	unsigned long* solve_energy();
	bool* solve_decision();
	t_w_arc* compute_winning_strategy(unsigned int p);
	t_w_arc* compute_optimal_strategy();
};

#endif
